from enum import Enum


class BaseType(Enum):
    PR = 'PR'
    FT = 'FT'


class DB_KEYS:
    class HARDWARE:
        @staticmethod
        def KEY() -> str:
            return "hardware"
        NAME = "name"
        IS_ENABLED = "is_enabled"

        class DEFINITION:
            @staticmethod
            def KEY() -> str:
                return "definition"
            TESTER = 'tester'

            class PCB:
                @staticmethod
                def KEY() -> str:
                    return "PCB"
                SINGLE_SITE_LOADBOARD = "SingleSiteLoadboard"
                SINGLE_SITE_DIB = "SingleSiteDIB"
                SINGLE_SITE_DIB_IS_CABLE = "SingleSiteDIBisCable"
                SINGLE_SITE_PROBE_CARD = "SingleSiteProbeCard"
                MULTI_SITE_LOADBOARD = "MultiSiteLoadboard"
                MULTI_SITE_DIB = "MultiSiteDIB"
                MULTI_SITE_DIB_IS_CABLE = "MultiSiteDIBisCable"
                MULTI_SITE_CARD = "MultiSiteProbeCard"
                MAX_PARALLELISM = "MaxParallelism"

            class PARALLELISM:
                @staticmethod
                def KEY() -> str:
                    return "Parallelism"
                NAME = "name"
                SITES = "sites"
                NEXT_PING_PONG_ID = "next_ping_pong_id"

                class CONFIGS:
                    @staticmethod
                    def KEY() -> str:
                        return "configs"
                    ID = "id"
                    NAME = "name"
                    STAGES = "stages"

            class ACTUATOR:
                @staticmethod
                def KEY() -> str:
                    return 'Actuator'
                PR = BaseType.PR.value
                FT = BaseType.FT.value

            class INSTRUMENTS:
                @staticmethod
                def KEY() -> str:
                    return "Instruments"

            class GP_FUNCTIONS:
                @staticmethod
                def KEY() -> str:
                    return 'GPFunctions'

    class SEQUENCE:
        OWNER_NAME = 'owner_name'
        PROG_NAME = 'prog_name'
        TEST_NAME = 'test'
        TEST_ORDER = 'test_order'

        class DEFINITION:
            @staticmethod
            def KEY():
                return 'definition'
            IS_SELECTED = "is_selected"
            DESCRIPTION = "description"
            NAME = "name"
            SBIN = "sbin"
            TEST_NUM = "test_num"
            EXECUTIONS = "executions"

            class INPUT_PARAMETERS:
                @staticmethod
                def KEY():
                    return "input_parameters"

                class TEMPERATURE:
                    @staticmethod
                    def KEY():
                        return "Temperature"
                    VALUE = "value"
                    TYPE = "type"
                    MIN = "Min"
                    MAX = "Max"
                    FMT = "fmt"
                    UNIT = "Unit"
                    EXPONENT = "10\u1d61"
                    DEFAULT = "Default"
                    CONTENT = "content"

            class OUTPUT_PARAMETERS:
                @staticmethod
                def KEY():
                    return "output_parameters"
                UTL = "UTL"
                LTL = "LTL"
                LSL = "LSL"
                USL = "USL"
                FMT = "fmt"
                UNIT = "Unit"
                EXPONENT = "10\u1d61"
                TEST_NUM = "test_num"

                class BINNING:
                    @staticmethod
                    def KEY():
                        return "Binning"
                    BIN = "bin"
                    RESULT = "result"
                    NAME = "name"
                    GROUP = "group"
                    DESCRIPTIO = "description"



from typing import Dict, List
from ate_projectdatabase.Types import Types
from ate_projectdatabase.FileOperator import DBObject, FileOperator
from uuid import uuid1

ExecutionSequenceType = Dict[str, List[int]]


class Program:

    @staticmethod
    def add(
        session: FileOperator, name: str, hardware: str, base: str, target: str, usertext: str,
        sequencer_typ: str, temperature: str, owner_name: str, order: int, cache_type: str,
        caching_policy: str, test_ranges: list, instance_count: int, execution_sequence: ExecutionSequenceType
    ):
        prog = {
            "id": str(uuid1()), "prog_name": name, "hardware": hardware, "base": base, "target": target, "usertext": usertext,
            "sequencer_type": sequencer_typ, "temperature": temperature, "owner_name": owner_name, "prog_order": order,
            "is_valid": True, "cache_type": cache_type, "caching_policy": caching_policy, "test_ranges": test_ranges,
            "instance_count": instance_count, "execution_sequence": execution_sequence
        }
        session.query(Types.Program()).add(prog)
        session.commit()

    @staticmethod
    def update(
        session: FileOperator, name: str, hardware: str, base: str, target: str, usertext: str,
        sequencer_type: str, temperature: str, owner_name: str, cache_type: str, caching_policy: str,
        test_ranges: list, instance_count: int, execution_sequence: ExecutionSequenceType
    ):
        prog = Program.get_by_name_and_owner(session, name, owner_name)
        prog.hardware = hardware
        prog.base = base
        prog.target = target
        prog.usertext = usertext
        prog.sequencer_type = sequencer_type
        prog.temperature = temperature
        prog.cache_type = cache_type
        prog.caching_policy = caching_policy
        prog.test_ranges = test_ranges
        prog.instance_count = instance_count
        prog.execution_sequence = execution_sequence
        session.commit()

    @staticmethod
    def remove(session: FileOperator, program_name: str, owner_name: str):
        session.query(Types.Program())\
               .filter(lambda Program: (Program.prog_name == program_name and Program.owner_name == owner_name))\
               .delete()
        session.commit()

    @staticmethod
    def get(session: FileOperator, name: str) -> DBObject:
        return session.query(Types.Program())\
                      .filter(lambda Program: Program.prog_name == name)\
                      .one()

    @staticmethod
    def get_by_name_and_owner(session: FileOperator, prog_name: str, owner_name: str) -> DBObject:
        return session.query(Types.Program())\
                      .filter(lambda Program: (Program.prog_name == prog_name and Program.owner_name == owner_name))\
                      .one()

    @staticmethod
    def get_by_order_and_owner(session: FileOperator, prog_order: str, owner_name: str) -> DBObject:
        return session.query(Types.Program())\
                      .filter(lambda Program: (Program.prog_order == prog_order and Program.owner_name == owner_name))\
                      .one()

    @staticmethod
    def update_program_name(session: FileOperator, prog_name: str, new_name: str):
        prog = Program.get(session, prog_name)
        prog.prog_name = new_name
        session.commit()

    @staticmethod
    def _update_program_order_neighbour(session: FileOperator, owner_name: str, prev_order: int, order: int, new_name: str, id: int):
        prog = session.query(Types.Program())\
                      .filter(lambda Program: (Program.owner_name == owner_name and Program.prog_order == prev_order and Program.id != id))\
                      .one()
        prog.prog_order = order
        prog.prog_name = new_name
        session.commit()

    @staticmethod
    def _update_program_order(session: FileOperator, owner_name: str, prev_order: int, order: int, new_name: str):
        prog = session.query(Types.Program())\
                      .filter(lambda Program: (Program.owner_name == owner_name and Program.prog_order == prev_order))\
                      .one()
        prog.prog_name = new_name
        prog.prog_order = order
        session.commit()

    def update_program_order(session: FileOperator, program_name: str, owner_name: str, order: int):
        program = session.query(Types.Program())\
                         .filter(lambda Program: Program.prog_name == program_name and Program.owner_name == owner_name)\
                         .one()
        program.prog_order = order
        session.commit()

    @staticmethod
    def get_programs_for_owner(session: FileOperator, owner_name: str) -> list:
        return session.query(Types.Program())\
                      .filter(lambda Program: Program.owner_name == owner_name).sort(lambda Program: Program.prog_order)\
                      .all()

    @staticmethod
    def get_all(session: FileOperator) -> List[DBObject]:
        return session.query(Types.Program())\
                      .all()

    @staticmethod
    def get_program_owner_element_count(session: FileOperator, owner_name: str) -> int:
        return session.query(Types.Program())\
                      .filter(lambda Program: Program.owner_name == owner_name)\
                      .count()

    @staticmethod
    def get_programs_for_hardware(session: FileOperator, hardware: str) -> List[DBObject]:
        return session.query(Types.Program())\
                      .filter(lambda Program: Program.hardware == hardware)\
                      .sort(lambda Program: Program.prog_order)\
                      .all()

    @staticmethod
    def update_program_order(session: FileOperator, new_order: int, owner_name: str, current_order: int):
        prog = Program.get_by_order_and_owner(session, current_order, owner_name)
        prog.prog_order = new_order
        session.commit()

    @staticmethod
    def get_programs_for_target(session: FileOperator, target_name: str) -> list:
        return session.query(Types.Program())\
                      .filter(lambda Program: (Program.target == target_name))\
                      .all()

    @staticmethod
    def get_program_execution_sequence(session: FileOperator, program_name: str) -> ExecutionSequenceType:
        cur_obj = Program.get(session, program_name)
        return cur_obj.execution_sequence

    @staticmethod
    def get_ping_pong_in_executions(session: FileOperator, parallelism_name: str, ping_pong_id: int) -> List[str]:
        ret = []
        for x in Program.get_all(session):
            if ping_pong_id in x.execution_sequence[parallelism_name]:
                ret.append(x.name)
        return ret

    @staticmethod
    def add_parallelism_to_execution_sequence(session: FileOperator, hardware: str, parallelism_name: str, ping_pong_id: int):
        for x in Program.get_programs_for_hardware(session, hardware):
            x.execution_sequence[parallelism_name] = [ping_pong_id for _ in range(x.instance_count)]
        session.commit()

    @staticmethod
    def remove_parallelism_from_execution_sequence(session: FileOperator, hardware: str, parallelism_name: str):
        for x in Program.get_programs_for_hardware(session, hardware):
            del x.execution_sequence[parallelism_name]
        session.commit()

    @staticmethod
    def set_program_validity(session: FileOperator, name: str, is_valid: bool):
        program = Program.get(session, name)
        program.is_valid = is_valid
        session.commit()




from typing import Dict, List, Optional, Tuple

from ate_projectdatabase.Hardware.ParallelismConfig import ParallelismConfig

from ate_projectdatabase.Utils import BaseType


class ParallelismStore:
    def __init__(self):
        self.parallelism_store: Dict[str, ParallelismConfig] = {}

    @staticmethod
    def from_database(parallelism_data: Dict[str, any]) -> "ParallelismStore":
        obj = ParallelismStore()
        for base_type, configs in parallelism_data.items():
            for table in configs:
                obj.add(ParallelismConfig.from_database(base_type, table))
        return obj

    def serialize(self) -> Dict[str, List[any]]:
        serialize = {
            BaseType.PR.value: [],
            BaseType.FT.value: []
        }
        for config_table in self.parallelism_store.values():
            if config_table.base_type == BaseType.PR:
                serialize[BaseType.PR.value].append(config_table.serialize())
            elif config_table.base_type == BaseType.FT:
                serialize[BaseType.FT.value].append(config_table.serialize())
        return serialize

    def add(self, config: ParallelismConfig):
        self.parallelism_store[config.name] = config

    def remove(self, name: str):
        self.parallelism_store.pop(name)

    def get(self, name: str) -> ParallelismConfig:
        return self.parallelism_store[name]

    def add_all(self, new_data: List[ParallelismConfig]):
        for table_obj in new_data:
            self.add(table_obj)

    def get_all_matching_base(self, base_type: BaseType) -> Dict[str, ParallelismConfig]:
        return {
            name: config for name, config in self.parallelism_store.items()
            if config.base_type == BaseType(base_type)
        }

    def get_all(self) -> Dict[str, ParallelismConfig]:
        return self.parallelism_store

    def get_count_matching_base(self, base_type: BaseType) -> int:
        return len(self.get_all_matching_base(base_type).keys())

    def all_tables_filled(self) -> bool:
        for table in self.parallelism_store.values():
            if not table.are_all_sites_used():
                return False
        return True

    @staticmethod
    def gen_suffix(num) -> str:
        letter_offset = ord('A')
        suffix = ''
        while num >= 26:
            letter = num % 26
            num = num // 26 - 1
            suffix += chr(letter_offset + letter)
        suffix += chr(letter_offset + num)
        return suffix[::-1]

    def generate_next_config_name(self, base_type: BaseType, add_sites_count: int):
        new_name = base_type.value + str(add_sites_count)
        name_suffix_num = 0
        name_suffix = ParallelismStore.gen_suffix(name_suffix_num)
        while str(new_name + name_suffix) in self.parallelism_store:
            name_suffix_num += 1
            name_suffix = ParallelismStore.gen_suffix(name_suffix_num)
        new_name += name_suffix
        return new_name

    def find_duplicate(self) -> Tuple[bool, Optional[str], Optional[str]]:
        index_pairs = [(a, b) for a in range(len(self.parallelism_store)) for b in range(a)]
        key_list = list(self.parallelism_store.keys())
        for (a, b) in index_pairs:
            if ParallelismStore.does_pattern_match(
                self.parallelism_store[key_list[a]],
                self.parallelism_store[key_list[b]]
            ):
                return True, key_list[a], key_list[b]
        return False, None, None

    @staticmethod
    def does_pattern_match(table_1: ParallelismConfig, table_2: ParallelismConfig) -> bool:
        if (len(table_1.cells) != len(table_2.cells)
                or table_1.base_type != table_2.base_type):
            return False
        if table_1 == table_2:
            return True
        for site_num, coord in table_1.cells.items():
            if table_2.cells[site_num] != coord:
                return False
        return True

    def min_required_parallelism(self) -> int:
        value = 0
        for table in self.parallelism_store.values():
            if table.sites_count > value:
                value = table.sites_count
        return value



from typing import Dict, List, Set

from ate_projectdatabase.Utils import DB_KEYS


class PingPong:
    def __init__(self, name: str, id: int, stage_count: int):
        self.id: int = id
        self.name: str = name
        self.stages: List[PingPongStage] = []
        self.stage_count: int = stage_count

    @staticmethod
    def from_database(data: Dict[str, any]) -> "PingPong":
        """The dict should contain the following keys:
        DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.CONFIGS.NAME with value of type "str"
        DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.CONFIGS.STAGES with value of type "typing.List[typing.List[int]"
        """
        obj = PingPong(
            data[DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.CONFIGS.NAME],
            data[DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.CONFIGS.ID],
            0
        )
        for stage_data in data[DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.CONFIGS.STAGES]:
            obj.stages.append(PingPongStage.from_database(stage_data))
        return obj

    def serialize(self) -> List[List[int]]:
        return {
            DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.CONFIGS.ID: self.id,
            DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.CONFIGS.NAME: self.name,
            DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.CONFIGS.STAGES: [
                ping_pong_stage.serialize() for ping_pong_stage in self.stages
            ]
        }

    @property
    def stage_count(self) -> int:
        return len(self.stages)

    @stage_count.setter
    def stage_count(self, value: int):
        if value > self.stage_count:
            for _ in range(value - self.stage_count):
                self.stages.append(PingPongStage())
        elif value < self.stage_count:
            self.stages = self.stages[:value]

    def is_site_used(self, site_num):
        for st in self.stages:
            if site_num in st.stage:
                return True
        return False


class PingPongStage:
    def __init__(self):
        self.stage: Set[int] = set()

    @staticmethod
    def from_database(stage: List[str]) -> "PingPongStage":
        obj = PingPongStage()
        obj.stage = set([int(elem) for elem in stage])
        return obj

    def serialize(self) -> List[str]:
        return list([str(elem) for elem in self.stage])



from ate_projectdatabase.Types import Types
from ate_projectdatabase.FileOperator import DBObject, FileOperator


class Maskset:

    @staticmethod
    def get(session: FileOperator, name: str) -> DBObject:
        return session.query(Types.Maskset())\
                      .filter(lambda Maskset: Maskset.name == name)\
                      .one()

    @staticmethod
    def get_all(session: FileOperator) -> list:
        return session.query(Types.Maskset()).all()

    @staticmethod
    def add(session: FileOperator, name: str, customer: str, definition: dict, is_enabled: bool):
        maskset = {"name": name, "customer": customer, "definition": definition, "is_enabled": is_enabled}
        session.query(Types.Maskset()).add(maskset)
        session.commit()

    @staticmethod
    def update(session: FileOperator, name: str, customer: str, definition: dict):
        maskset = Maskset.get(session, name)
        maskset.definition = definition
        maskset.customer = customer
        session.commit()

    @staticmethod
    def update_state(session: FileOperator, name: str, is_enabled: bool):
        maskset = Maskset.get(session, name)
        maskset.is_enabled = is_enabled
        session.commit()

    @staticmethod
    def remove(session: FileOperator, name: str):
        session.query(Types.Maskset())\
               .filter(lambda Maskset: Maskset.name == name)\
               .delete()
        session.commit()

    @staticmethod
    def get_definition(session: FileOperator, name: str) -> dict:
        maskset = Maskset.get(session, name)
        return maskset.definition

    @staticmethod
    def get_customer(session: FileOperator, name: str) -> str:
        maskset = Maskset.get(session, name)
        return maskset.customer

    @staticmethod
    def get_ASIC_masksets(session: FileOperator) -> list:
        return session.query(Types.Maskset())\
                      .filter(lambda Maskset: Maskset.customer != '')\
                      .all()

    @staticmethod
    def get_ASSP_masksets(session: FileOperator) -> list:
        return session.query(Types.Maskset())\
                      .filter(lambda Maskset: Maskset.customer == '')\
                      .all()



from ate_projectdatabase.Types import Types
from ate_projectdatabase.FileOperator import DBObject, FileOperator


class Test:

    @staticmethod
    def get(session: FileOperator, name: str, hardware: str, base: str) -> DBObject:
        return session.query(Types.Test())\
                      .filter(lambda Test: (Test.name == name and Test.hardware == hardware and Test.base == base))\
                      .one()

    @staticmethod
    def get_one_or_none(session: FileOperator, name: str, hardware: str, base: str) -> DBObject:
        return session.query(Types.Test())\
                      .filter(lambda Test: (Test.name == name and Test.hardware == hardware and Test.base == base))\
                      .one_or_none()

    @staticmethod
    def get_for_hw_base_test_typ(session: FileOperator, hardware: str, base: str, test_type: str) -> list:
        if test_type != 'all':
            return session.query(Types.Test())\
                          .filter(lambda Test: (Test.base == base and Test.hardware == hardware and Test.type == test_type))\
                          .all()
        else:
            return session.query(Types.Test())\
                          .filter(lambda Test: (Test.base == base and Test.hardware == hardware))\
                          .all()

    @staticmethod
    def get_all(session: FileOperator) -> list:
        return session.query(Types.Test())\
                      .all()

    @staticmethod
    def remove(session: FileOperator, name: str, hardware: str, base: str):
        session.query(Types.Test())\
               .filter(lambda Test: Test.name == name and Test.hardware == hardware and Test.base == base)\
               .delete()
        session.commit()

    @staticmethod
    def replace(session: FileOperator, database):
        name = database['name']
        hardware = database['hardware']
        base = database['base']
        test_type = database['type']
        session.query(Types.Test())\
               .filter(lambda Test: Test.name == name and Test.hardware == hardware and Test.base == base)\
               .delete()

        test = {"name": name, "hardware": hardware, "base": base, "type": test_type, "definition": database, "is_enabled": True}
        session.query(Types.Test()).add(test)

        session.commit()

    @staticmethod
    def update(session: FileOperator, name: str, hardware: str, base: str, type: str, definition: dict, is_enabled: bool):
        test = session.query(Types.Test())\
                      .filter(lambda Test: (Test.name == name and Test.hardware == hardware and Test.base == base))\
                      .one()
        test.definition = definition
        test.is_enabled = is_enabled
        session.commit()

    @staticmethod
    def add(session: FileOperator, name: str, hardware: str, base: str, test_type: str, definition: dict, is_enabled: bool):
        test = {"name": name, "hardware": hardware, "base": base, "type": test_type, "definition": definition, "is_enabled": is_enabled}
        session.query(Types.Test()).add(test)
        session.commit()

    @staticmethod
    def get_all_for_hardware(session: FileOperator, hardware: str) -> list:
        return session.query(Types.Test()).filter(lambda Test: Test.hardware == hardware).all()



from typing import Dict, List, Optional, Tuple

from ate_projectdatabase.Hardware.PingPong import PingPong, PingPongStage
from ate_projectdatabase.Utils import DB_KEYS, BaseType


DEFAULT_FIRST_PING_PONG_NAME = "All Parallel"


class ParallelismConfig:
    def __init__(self):
        self.name: str = None
        self.base_type: BaseType = None
        self.sites_count: int = None
        self.cells: Dict[int, Tuple(int, int)] = {}
        self.configs: List[PingPong[PingPongStage]] = []
        self.next_ping_pong_id: int = 0

    @staticmethod
    def new(name: str, base_type: BaseType, sites_count: int):
        obj = ParallelismConfig()
        obj.name = name
        assert isinstance(base_type, BaseType)
        obj.base_type = base_type
        obj.sites_count = sites_count
        obj._add_default_first_ping_pong()
        return obj

    def _add_default_first_ping_pong(self):
        self.add_ping_pong_config(DEFAULT_FIRST_PING_PONG_NAME, 1)
        self.configs[0].stages[0].stage = [site for site in range(self.sites_count)]

    @staticmethod
    def from_database(base_type: BaseType, data: Dict[str, any]) -> "ParallelismConfig":
        obj = ParallelismConfig()
        obj.name = data[DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.NAME]
        obj.base_type = BaseType(base_type)
        obj.sites_count = len(data[DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.SITES])

        for index, coord in enumerate(data[DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.SITES]):
            obj.cells[index] = tuple(coord)
        for new_ping_pong in data[DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.CONFIGS.KEY()]:
            obj.configs.append(PingPong.from_database(new_ping_pong))
        obj.next_ping_pong_id = data[DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.NEXT_PING_PONG_ID]
        return obj

    def serialize(self) -> Dict[str, any]:
        return {
            DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.NEXT_PING_PONG_ID: self.next_ping_pong_id,
            DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.NAME: self.name,
            DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.SITES: list(self.cells.values()),
            DB_KEYS.HARDWARE.DEFINITION.PARALLELISM.CONFIGS.KEY(): [ping_pong.serialize() for ping_pong in self.configs]
        }

    def are_all_sites_used(self) -> bool:
        return self.sites_count == len(self.cells)

    def add_ping_pong_config(self, name: str, stage_count: int):
        self.configs.append(PingPong(name, self.next_ping_pong_id, stage_count))
        self.next_ping_pong_id += 1

    def remove_ping_pong_config(self, rem_item: PingPong):
        if rem_item in self.configs:
            self.configs.remove(rem_item)

    def get_default_first_ping_pong(self) -> PingPong:
        if len(self.configs) == 0:
            self._add_default_first_ping_pong()
        return self.configs[0]

    def get_all_ping_pong_names(self) -> List[str]:
        return [ping_pong.name for ping_pong in self.configs]

    def get_ping_pong(self, name: str) -> Optional[PingPong]:
        for ping_pong in self.configs:
            if name == ping_pong.name:
                return ping_pong
        return None

    def get_ping_pong_by_id(self, id: int) -> Optional[PingPong]:
        for ping_pong in self.configs:
            if id == ping_pong.id:
                return ping_pong
        return None

    def are_all_configs_correct(self) -> Tuple[bool, Optional[str]]:
        for ping_pong in self.configs:
            all_sites = set()
            for st in ping_pong.stages:
                if not st.stage:
                    return False, f'{self.name}.{ping_pong.name} has an empty stage.'
                if all_sites.intersection(st.stage):
                    return False, f'{self.name}.{ping_pong.name}: A site can only be in one stage!'
                all_sites.update(st.stage)
            if len(all_sites) != self.sites_count:
                return False, f'{self.name}.{ping_pong.name} there are not all sites used.'
        return True, None



from typing import Callable
import os
import json

from ate_projectdatabase.Types import Types

CONFIG_DIR_NAME = 'definitions'


class DBObject:
    def to_dict(self) -> dict:
        return self.__dict__

    def write_attribute(self, member_name: str, value):
        self.__dict__[member_name] = value

    def read_attribute(self, attribute_name: str):       # Returntype can be just about anything!
        return self.__dict__[attribute_name]

    def has_attribute(self, attribute_name: str) -> bool:
        return attribute_name in self.__dict__


class FileOperator:
    def __init__(self, project_dir: str):
        self.config_dir = os.path.join(project_dir, CONFIG_DIR_NAME)
        self.data_cache = {}
        self.query_open = False
        self.filter_expression = lambda x: True
        self.sort_expression = None
        self.current_type = None
        self.current_subtypes = []

    @staticmethod
    def _make_db_object(json_dict: dict) -> DBObject:
        current_object = DBObject()
        for k, v in json_dict.items():
            setattr(current_object, k, v)
        return current_object

    def get_current_target_list_name(self) -> str:
        '''
            Tricky: A query can concern multiple files, however,
            if we want to add an item it shall only ever be added
            to a single file. To find out, which file that is, we
            build the path pattern and take the first file, that
            matches, throwing an exception if more than one file
            matches and generating a new filename if no name file
            matches
        '''
        file_path = self.generate_path(self.current_type, self.current_subtypes)
        import glob
        if len(glob.glob(file_path)) > 1:
            raise IOError("Tried to add data to multiple files")
        if len(glob.glob(file_path)) == 1:
            return glob.glob(file_path).pop(0)

        # Generate new filename, as apparently no file exists...
        the_type = self.generate_path_base(self.current_type, self.current_subtypes)
        file_path = os.path.join(self.config_dir, self.current_type, f'{the_type}.json')
        self.data_cache[file_path] = []
        return file_path

    def load_configuration(self, type: Types, subtypes: list):
        file_path = self.generate_path(type, subtypes)
        self.data_cache = {}

        import glob
        for f in glob.glob(file_path):
            try:
                with open(f, 'r') as f:
                    loaded_data = json.load(f)
                    return_val = []
                    for item in loaded_data:
                        return_val.append(self._make_db_object(item))
                self.data_cache[f.name] = return_val
            except Exception:
                self.data_cache[f.name] = []

        # nastyness: if nothing was found in the FS we atleast create
        # the file thet specifies our filename:
        if len(self.data_cache) == 0:
            file_path = file_path.replace("*", "")
            self.data_cache[file_path] = []

    def store_configuration(self):
        # DBObjects are not easily serializable, we serialize
        # them each on their own...
        for name, itemlist in self.data_cache.items():
            print(f'Name: {name}')
            data_to_write = []
            for item in itemlist:
                data_to_write.append(item.__dict__)

            if len(data_to_write) > 0:
                with open(name, 'w') as f:
                    json.dump(data_to_write, f, indent=2)
            else:
                # delete file if still there,
                # but no records are available for the file.
                if os.path.exists(name):
                    os.remove(name)

    def generate_path_base(self, type: Types, subtypes: list) -> str:
        the_type = type
        for subtype in subtypes:
            the_type = the_type + subtype
        return the_type

    def generate_path(self, type: Types, subtypes: list) -> str:
        self.current_type = type
        self.current_subtypes = subtypes

        the_type = self.generate_path_base(type, subtypes)

        file_path = os.path.join(self.config_dir, type, f'{the_type}*.json')
        return file_path

    def query(self, type: Types):
        # self.data_cache contains the complete representation of the available data and is
        # always current provided no two instances of the FileOperator work on the
        # same dataset. This means we can avoid hitting masstorage for loading if
        # the dataset (i.e. type!) did not change between two consecutive calls
        # to query.
        self.query_with_subtype(type, [])
        return self

    def query_with_subtype(self, type: Types, subtypes: list):
        if type != self.current_type or subtypes != self.current_subtypes:
            self.load_configuration(type, subtypes)
            self.current_type = type
            self.current_subtypes = subtypes
        self.query_open = True
        self.filter_expression = lambda x: True
        self.sort_expression = None
        return self

    def filter(self, filter_expression: Callable):
        if self.query_open is False:
            raise Exception("Cannot apply filter when no query is open.")
        self.filter_expression = filter_expression
        return self

    def sort(self, sort_expression: Callable):
        if self.query_open is False:
            raise Exception("Cannot sort when no query is open.")
        self.sort_expression = sort_expression
        return self

    def all(self) -> list:
        if self.query_open is False:
            raise Exception("Cannot get 'all' when no query is open.")
        data = list(x for x in self._items() if self.filter_expression(x))
        if self.sort_expression is not None:
            data = sorted(data, key=self.sort_expression)
        return data

    def one(self) -> DBObject:
        if self.query_open is False:
            raise Exception("Cannot get 'one' when no query is open.")
        items = self.all()
        assert (len(items) == 1)
        return items[0]

    def one_or_none(self) -> DBObject:
        if self.query_open is False:
            raise Exception("Cannot get 'one or none' when no query is open.")
        items = self.all()
        if len(items) == 0:
            return None
        # If we have items, having more than one is an error in this case!
        assert (len(items) == 1)
        return items[0]

    def _items(self):
        import itertools
        all_loaded_items = [y for _, y in self.data_cache.items()]
        all_loaded_items = [x for x in itertools.chain(*all_loaded_items)]
        return all_loaded_items

    def delete(self):
        if self.query_open is False:
            raise Exception("Cannot 'delete' when no query is open.")
        for f, items in self.data_cache.items():
            new_list = list(x for x in items if self.filter_expression(x) is False)
            self.data_cache[f] = new_list
        return self

    def delete_item(self, item: DBObject):
        if self.query_open is False:
            raise Exception("Cannot 'delete_item' when no query is open.")
        self.data_cache.pop(item)
        return self

    def insert(self, objects: list):
        if self.query_open is False:
            raise Exception("Cannot 'insert'' when no query is open.")
        for item in objects:
            for f, itemlist in self.data_cache.items():
                itemlist.append(self._make_db_object(item))
        return self

    def add(self, object: dict):
        if self.query_open is False:
            raise Exception("Cannot 'insert' when no query is open.")
        list_name = self.get_current_target_list_name()
        self.data_cache[list_name].append(self._make_db_object(object))
        return self

    def count(self) -> int:
        if self.query_open is False:
            raise Exception("Cannot 'count' when no query is open.")
        return len(self.all())

    def commit(self):
        if self.query_open is False:
            raise Exception("Cannot 'commit' when no query is open.")
        self.store_configuration()
        self.query_open = False

    def rename(self, type: Types, old_name: str, new_name: str):
        base_path = os.path.join(self.config_dir, self.generate_path_base(type, []))
        old_name = f"{type}{old_name}.json"
        new_name = f"{type}{new_name}.json"
        old_path = os.path.join(base_path, old_name)
        new_path = os.path.join(base_path, new_name)
        os.rename(old_path, new_path)



from ate_projectdatabase.Types import Types
from ate_projectdatabase.FileOperator import DBObject, FileOperator


class Version:
    @staticmethod
    def get(session: FileOperator) -> DBObject:
        return session.query(Types.Version())\
                      .one()



from ate_projectdatabase.Types import Types
from ate_projectdatabase.FileOperator import FileOperator


class Sequence:

    @staticmethod
    def add_sequence_information(session: FileOperator, owner_name: str, prog_name: str, test: str, order: int, definition: dict):
        s = {"owner_name": owner_name, "prog_name": prog_name, "test": test, "test_order": order, "definition": definition}
        session.query_with_subtype(Types.Sequence(), prog_name).add(s)
        session.commit()

    @staticmethod
    def get_for_program(session: FileOperator, prog_name: str) -> list:
        return session.query_with_subtype(Types.Sequence(), prog_name)\
                      .filter(lambda Sequence: Sequence.prog_name == prog_name)\
                      .sort(lambda Sequence: Sequence.test_order)\
                      .all()

    @staticmethod
    def get_programs_for_test(session: FileOperator, test_name: str) -> list:
        return session.query(Types.Sequence())\
                      .filter(lambda Sequence: Sequence.test == test_name)\
                      .all()

    @staticmethod
    def remove_test_from_sequence(session: FileOperator, test_name: str):
        session.query(Types.Sequence())\
               .filter(lambda Sequence: Sequence.test == test_name)\
               .delete()
        session.commit()

    @staticmethod
    def remove_program_sequence(session: FileOperator, prog_name: str, owner_name: str):
        session.query(Types.Sequence())\
               .filter(lambda Sequence: (Sequence.prog_name == prog_name and Sequence.owner_name == owner_name))\
               .delete()
        session.commit()

    @staticmethod
    def remove_for_program(session: FileOperator, program_name: str):
        session.query_with_subtype(Types.Sequence(), program_name)\
               .filter(lambda Sequence: Sequence.prog_name == program_name)\
               .delete()
        session.commit()

    @staticmethod
    def update_progname(session: FileOperator, old_prog_name: str, new_prog_name: str):
        progs = Sequence.get_for_program(session, old_prog_name)
        for p in progs:
            p.prog_name = new_prog_name
        session.commit()

        # Hacky: Since this used to be done by means of sqlite which would
        # just delete everything it finds, we eat the exceptions here...
        try:
            session.rename(Types.Sequence(), old_prog_name, f"{new_prog_name}_tmp")
        except:
            pass
        try:
            session.rename(Types.Sequence(), new_prog_name, f"{old_prog_name}_tmp")
        except:
            pass
        try:
            session.rename(Types.Sequence(), f"{new_prog_name}_tmp", new_prog_name)
        except:
            pass
        try:
            session.rename(Types.Sequence(), f"{old_prog_name}_tmp", old_prog_name)
        except:
            pass

    @staticmethod
    def switch_sequences(session: FileOperator, prog_name: str, new_prog_name: str):
        Sequence.update_progname(session, prog_name, new_prog_name)

    @staticmethod
    def remove(session: FileOperator, program_name: str, owner_name: str, program_order: str):
        session.query_with_subtype(Types.Sequence(), program_name)\
               .filter(lambda Sequence: (Sequence.prog_name == program_name and Sequence.owner_name == owner_name))\
               .delete()
        session.commit()



from ate_projectdatabase.Types import Types
from ate_projectdatabase.FileOperator import DBObject, FileOperator


class Group:
    @staticmethod
    def get(session: FileOperator, name: str) -> DBObject:
        return session.query(Types.Group())\
                      .filter(lambda Group: Group.name == name)\
                      .one()

    @staticmethod
    def get_all(session: FileOperator) -> DBObject:
        return session.query(Types.Group()).all()

    @staticmethod
    def add(session: FileOperator, name: str, is_standard: bool = False):
        group = {"name": name, "is_selected": True, "is_standard": is_standard, "programs": [], "tests": []}
        session.query(Types.Group()).add(group)
        session.commit()

    @staticmethod
    def update_state(session: FileOperator, name: str, is_selected: bool):
        group = session.query(Types.Group()).filter(lambda Group: Group.name == name).one()
        group.is_selected = is_selected
        session.commit()

    @staticmethod
    def remove(session: FileOperator, name: str):
        session.query(Types.Group())\
               .filter(lambda Group: Group.name == name)\
               .delete()
        session.commit()

    @staticmethod
    def is_standard(session: FileOperator, name: str) -> bool:
        group = session.query(Types.Group()).filter(lambda Group: Group.name == name).one()
        return group.is_standard

    @staticmethod
    def add_testprogram_to_group(session: FileOperator, group: str, prog_name: str):
        group = session.query(Types.Group()).filter(lambda Group: Group.name == group).one()
        group.programs.append(prog_name)
        session.commit()

    @staticmethod
    def remove_testprogram_from_group(session: FileOperator, group: str, prog_name: str):
        group = session.query(Types.Group()).filter(lambda Group: Group.name == group).one()
        group.programs.pop(group.programs.index(prog_name))
        session.commit()

    @staticmethod
    def remove_test_from_group(session: FileOperator, group: str, test: str):
        group = session.query(Types.Group()).filter(lambda Group: Group.name == group).one()
        group.tests.pop(group.tests.index(test))
        session.commit()

    @staticmethod
    def get_programs_for_group(session: FileOperator, group: str) -> list:
        group = session.query(Types.Group()).filter(lambda Group: Group.name == group).one()
        return group.programs

    @staticmethod
    def get_all_groups_for_test(session: FileOperator, test_name: str):
        return session.query(Types.Group()).filter(lambda Group: test_name in Group.tests).all()

    @staticmethod
    def update_groups_for_test(session: FileOperator, test: str, groups: list):
        all_groups_contains_test = Group.get_all_groups_for_test(session, test)
        for group_name in groups:
            group = session.query(Types.Group()).filter(lambda Group: Group.name == group_name).one()
            if test in group.tests:
                continue

            group.tests.append(test)

        if [group.name for group in all_groups_contains_test] != groups:
            # remove from list if needed
            for group in all_groups_contains_test:
                if group.name in groups:
                    continue

                group.tests.pop(group.tests.index(test))

        session.commit()

    @staticmethod
    def add_test_to_group(session: FileOperator, group: str, test: str):
        group = session.query(Types.Group()).filter(lambda Group: Group.name == group).one()
        group.tests.append(test)
        session.commit()

    @staticmethod
    def get_tests_for_group(session: FileOperator, group: str):
        group = session.query(Types.Group()).filter(lambda Group: Group.name == group).one()
        return group.tests
